# MIT License
#
# Copyright (c) 2021 Zhang.H.N
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import Optional
import sklearn.metrics as metrics

import torch.optim as optim
import torch.nn as nn
import torch
import numpy as np
import pandas as pd
from os.path import join

from socube.utils.io import (checkExist, writeCsv, mkDirs, rm, loadTorchModule)
from socube.utils.memory import (visualBytes, autoClearIter)
from socube.utils.logging import (getJobId, log)
from socube.net import NetBase
from socube.train import (EarlyStopping, evaluateReport)
from .data import ConvClassifyDataset
from .model import SoCubeNet
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
"""A module provided for model training and evaluation"""
__all__ = ["fit", "validate", "infer"]


def fit(home_dir: str,
        data_dir: str,
        lr: float = 0.001,
        gamma: float = 0.99,
        epochs: int = 100,
        train_batch: int = 32,
        valid_batch: int = 500,
        transform: nn.Module = None,
        in_channels: int = 10,
        num_workers: int = 0,
        shuffle: bool = False,
        seed: int = None,
        label_file: str = "label.csv",
        threshold: float = 0.5,
        k: int = 5,
        once: bool = False,
        use_index: bool = True,
        device_name: str = "cpu",
        step: int = 5,
        model_id: str = None,
        pretrain_model_path: str = None,
        max_acc_limit: float = 1,
        **kwargs) -> str:
    """
    Train socube model.

    Parameters
    ----------
    home_dir: str
        the home directory of the specfic job
    data_dir: str
        the dataset's directory
    lr: float
        learning rate, default: 0.001
    gamma: float
        learning rate decay, default: 0.99
    epochs: int
        training epochs, default: 100
    train_batch: int
        training batch size, default: 32
    valid_batch: int
        validation batch size, default: 500
    transform: nn.Module
        sample transform, such as `Resize`
    in_channels: int
        the number of input channels, default: 10
    num_workers: int
        the number of workers for data loading, default: 0
    shuffle: bool
        if `True`, data will be shuffled while k-fold cross-valid
    seed: int
        random seed for k-fold cross-valid or sample
    label_file: str
        the label file csv name, default: "label.csv",
    threshold: float
        the threshold for classification, default: 0.5
    k: int
        k value of k-fold cross-valid
    once: bool
        if `True`, k-fold cross-validation runs first fold only
    use_index: bool
        If `True`, it will read sample file by index. Otherwise, it will read
        sample file by sample name.
    device_name: str
        the device name, default: "cpu"
    step: int
        the epoch step of learning rate decay, default: 5
    model_id: str
        the model id, If `None`, it will be generated automatically
    pretrain_model_path: str
        the pretrain model path, if not `None`, it will load the pretrain model
    max_acc_limit: float
        the max accuracy limit, if the accuracy is higher than this limit, the
        training will stop to prevent overfitting.
    **kwargs: dict
        the other parameters wanted to be saved in the log file.

    Returns
    ----------
    job id string
    """
    if seed is not None:
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    model_id = model_id if model_id else getJobId()
    log("Train", "Model ID: " + model_id)

    dataset = ConvClassifyDataset(data_dir=data_dir,
                                  transform=transform,
                                  labels=label_file,
                                  seed=seed,
                                  shuffle=shuffle,
                                  k=k,
                                  use_index=use_index)
    shape = dataset[0]["data"].shape
    mkDirs(f"{home_dir}/outputs/{model_id}")
    mkDirs(f"{home_dir}/models/{model_id}")
    mkDirs(f"{home_dir}/plots/{model_id}")
    report = pd.DataFrame()
    train_record = pd.DataFrame()
    kwargs.update({
        "job_id": model_id,
        "model": SoCubeNet.__name__,
        "channel": in_channels,
        "kFold": k,
        "lr": lr,
        "gamma": gamma,
        "transform": transform,
        "update_step": step,
        "train_batch": train_batch,
        "valid_batch": valid_batch,
        "epochs": epochs,
        "home_dir": home_dir,
        "pretrain_model": pretrain_model_path,
        "data_dir": data_dir,
        "random_state": seed,
        "label": label_file,
        "step": step,
        "onlyOne": once,
        "maxAccLimit": max_acc_limit
    })
    config = pd.Series(kwargs)
    device = torch.device(device_name if torch.cuda.is_available() else "cpu")

    # Specific the default gpu device by context manager
    # If not, pytorch with use cuda:0 as default and load data to it when intializing,
    # no matter what device you want to use.
    # If free memory of cuda:0 is less than required, a Runtime Error will occurred
    # Otherwise, you will find TWO gpu device are used by your process in `nvidia-smi` report.
    with torch.cuda.device(device if device.type == "cuda" else -1):
        for fold, (train_set,
                   valid_set) in autoClearIter(enumerate(dataset.kFold, 1)):
            data_loader = DataLoader(dataset=train_set,
                                     sampler=dataset.sampler(train_set),
                                     batch_size=train_batch,
                                     num_workers=num_workers)
            valid_loader = DataLoader(dataset=valid_set,
                                      batch_size=valid_batch,
                                      sampler=dataset.sampler(valid_set))

            # init and load pretrained model
            model = SoCubeNet(in_channels,
                              dataset.typeCounts,
                              freeze=pretrain_model_path,
                              binary=True,
                              shape=shape).to(device)

            log("Train", f"Shape is {model.shape}")
            if pretrain_model_path:
                mp = pretrain_model_path % (fold)
                log("Train", "Loading pretained model %s" % (mp))
                loadTorchModule(model, mp)

            optimizer = optim.Adam(model.parameters(), lr=lr)
            scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer,
                                                               gamma=gamma)

            valid_loss, valid_acc, valid_label, valid_score = validate(
                data_loader=valid_loader, model=model, device=device)

            loop = tqdm(range(epochs))
            train_loss_array = np.full(epochs, -1)
            valid_loss_array = np.full(epochs, -1)
            train_acc_array = np.full(epochs, -1)
            valid_acc_array = np.full(epochs, -1)

            earlystop = EarlyStopping(
                path=f"{home_dir}/models/{model_id}/modelTmp.pt",
                patience=10,
                verbose=10)

            for epoch in autoClearIter(loop):
                model.train()
                train_acc = 0
                train_loss = 0
                for index, batch in autoClearIter(enumerate(data_loader, 1)):
                    data, label = batch.values()
                    data = data.to(device)
                    label = label.to(device)
                    score = model(data)
                    loss = model.criterion(score, label)
                    optimizer.zero_grad()
                    loss.backward()
                    optimizer.step()
                    if model._binary:
                        predict = (score >= 0.5).cpu().long()
                    else:
                        predict = score.argmax(1).cpu()
                    train_acc += metrics.accuracy_score(label.cpu(), predict)
                    train_loss += loss.item()
                    loop.set_description("Fold %02d batch %03d/%03d" %
                                         (fold, index, len(data_loader)))
                    loop.set_postfix(
                        loss=train_loss / index,
                        train_acc=train_acc / index,
                        valid_acc=valid_acc,
                        best_loss=earlystop._best,
                        memo_res=visualBytes(
                            torch.cuda.memory_reserved(device)),
                        max_memo_res=visualBytes(
                            torch.cuda.max_memory_reserved(device)),
                        lr="%.06f" % (scheduler.get_last_lr()[0]))

                    del data, label, score, loss

                if (epoch + 1) % step == 0:
                    scheduler.step()

                valid_loss, valid_acc, valid_label, valid_score = validate(
                    data_loader=valid_loader, model=model, device=device)

                if earlystop(valid_loss, model, valid_acc < max_acc_limit):
                    best_label = valid_label
                    best_score = valid_score

                train_acc_array[epoch] = train_acc / index
                valid_acc_array[epoch] = valid_acc
                train_loss_array[epoch] = train_loss / index
                valid_loss_array[epoch] = valid_loss
                del valid_label, valid_score
                if earlystop.earlyStop:
                    log(__name__, "Early stopped for training")
                    break

            loadTorchModule(model, earlystop._path)
            model.cpu()
            torch.save(
                model.state_dict(),
                f"{home_dir}/models/{model_id}/{SoCubeNet.__name__}_{fold}.pt")
            rm(earlystop._path)
            evaluate_result = evaluateReport(
                best_label, best_score,
                f"{home_dir}/plots/{model_id}/{SoCubeNet.__name__}_ROC_{fold}.png",
                f"{home_dir}/plots/{model_id}/{SoCubeNet.__name__}_PRC_{fold}.png",
                threshold=threshold
            )

            log("Train", "Best validate ACC: %.4f" % (evaluate_result["ACC"]))
            del model

            report[f"fold_{fold}"] = evaluate_result
            if epochs > 0:
                train_record[f"fold_{fold}_train_loss"] = train_loss_array
                train_record[f"fold_{fold}_valid_loss"] = valid_loss_array
                train_record[f"fold_{fold}_train_ACC"] = train_acc_array
                train_record[f"fold_{fold}_valid_ACC"] = valid_acc_array

            # quit k-fold
            if once:
                break

    report["average"] = report.mean(axis=1)
    report["sample_stdev"] = report.std(axis=1)
    writeCsv(
        report["average"].append(config),
        f"{home_dir}/outputs/{model_id}/{SoCubeNet.__name__}_aveReport.csv",
        header=None)

    if epochs > 0:
        writeCsv(
            train_record,
            f"{home_dir}/outputs/{model_id}/{SoCubeNet.__name__}_trainRecord.csv",
            index=None)

    writeCsv(
        report,
        f"{home_dir}/outputs/{model_id}/{SoCubeNet.__name__}_kFoldReport.csv")

    return model_id


@torch.no_grad()
def validate(data_loader: DataLoader,
             model: NetBase,
             device: torch.device,
             with_progress: bool = False) -> tuple:
    """
    Validate model performance basically

    Parameters
    ----------
    dataLoader: the torch dataloader object used for validation
    model: Network model implemented `NetBase`
        the model waited for validation
    device: the cpu/gpu device

    Returns
    ----------
    a quadra tuple of (average loss, average ACC, true label, predict score)
    """
    with torch.cuda.device(device if device.type == "cuda" else -1):
        model.to(device)
        model.eval()
        loss_ave = 0
        acc_ave = 0
        score_list = list()
        label_list = list()

        itererate = autoClearIter(enumerate(data_loader, 1))
        if with_progress:
            itererate = tqdm(itererate, desc="Validate")

        for index, batch in itererate:
            data, label = batch.values()
            data = data.to(device)
            label = label.to(device)
            score = model(data)
            loss = model.criterion(score, label)
            if model._binary:
                predict = (score >= 0.5).cpu().long()
            else:
                predict = score.argmax(1).cpu()

            acc = metrics.accuracy_score(label.cpu(), predict)
            loss_ave += loss.cpu().item()
            acc_ave += acc
            score_list.extend(score.cpu().numpy())
            label_list.extend(label.cpu().numpy())

            # destroy useless object and recycle GPU memory
            del data, label, loss, score

        loss_ave /= index
        acc_ave /= index

    return loss_ave, acc_ave, np.array(label_list), np.array(score_list)


def infer(data_dir: str,
          home_dir: str,
          model_id: str,
          label_file: str,
          in_channels: int = 10,
          k: int = 5,
          threshold: float = 0.5,
          batch_size: int = 400,
          device_name: str = "cpu",
          with_eval: bool = False,
          seed: Optional[int] = None):
    device = torch.device(device_name if torch.cuda.is_available() else "cpu")
    dataset = ConvClassifyDataset(data_dir=data_dir,
                                  labels=label_file,
                                  shuffle=False,
                                  seed=seed,
                                  use_index=False)
    dataloader = DataLoader(dataset=dataset, batch_size=batch_size)
    log("Inference",
        f"Current threshold: {threshold}, batch_size: {batch_size}")
    model_dir = join(home_dir, "models", model_id)
    checkExist(model_dir, types="dir")
    output_dir = join(home_dir, "outputs", model_id)
    if with_eval:
        plot_dir = join(home_dir, "plots", model_id)
        mkDirs(plot_dir)
    mkDirs(output_dir)
    ensemble_score_list = []
    for i in range(1, k + 1):
        model = SoCubeNet(in_channels=in_channels, out_channels=2, binary=True)
        loadTorchModule(model,
                        join(model_dir, f"{SoCubeNet.__name__}_{i}.pt"),
                        skipped=False)
        _, _, label, score = validate(dataloader, model, device, True)
        assert (dataset._labels.iloc[:, 0] == label).all()
        if with_eval:
            writeCsv(
                evaluateReport(
                    label=label,
                    score=score,
                    roc_plot_file=join(plot_dir,
                                       f"inference_roc_{i}_{threshold}.png"),
                    prc_plot_file=join(plot_dir,
                                       f"inference_prc_{i}_{threshold}.png"),
                    threshold=threshold),
                join(output_dir, f"inference_report_{i}_{threshold}.csv"))

        writeCsv(pd.DataFrame({"score": score}, index=dataset._labels.index),
                 join(output_dir, f"inference_score_{i}.csv"))
        ensemble_score_list.append(score)

    log("Inference", "Model ensembling")
    ensemble_score = sum(ensemble_score_list) / k

    if with_eval:
        writeCsv(
            evaluateReport(
                label=label,
                score=ensemble_score,
                roc_plot_file=join(plot_dir, f"inference_roc_{threshold}.png"),
                prc_plot_file=join(plot_dir, f"inference_prc_{threshold}.png"),
                threshold=threshold),
            join(output_dir, f"inference_report_{threshold}.csv"))

    result = pd.DataFrame({"predict_score": ensemble_score},
                          index=dataset._labels.index)
    result["predict_type"] = result["predict_score"].apply(
        lambda s: "doublet" if s >= threshold else "singlet")
    writeCsv(result, join(output_dir, f"final_result_{threshold}.csv"))
